@@ -261,7 +261,7 @@ impl Language {
     /// _outside_ counts.
     fn uses(&self, request: &TypeSchema, expr: &Expression) -> (f64, Uses) {
         let mut ctx = Context::default();
-        let tp = request.instantiate(&mut ctx);
+        let tp = request.clone().instantiate_owned(&mut ctx);
         let env = Rc::new(LinkedList::default());
         self.likelihood_uses(&tp, expr, &ctx, &env)
     }
@@ -311,7 +311,7 @@ impl Language {
                         if expr != f {
                             continue;
                         }
-                    } else if let Some(frag_tp) =
+                    } else if let Some(mut frag_tp) =
                         TreeMatcher::do_match(self, ctx.to_mut(), expr, f, &mut bindings, xs.len())
                     {
                         let mut template: VecDeque<Type> =
@@ -327,7 +327,8 @@ impl Language {
                                 frag_tp,
                                 self.display(f),
                             ));
-                        tp = Cow::Owned(frag_tp.apply(&ctx));
+                        frag_tp.apply_mut(&ctx);
+                        tp = Cow::Owned(frag_tp);
                     } else {
                         continue;
                     }
@@ -355,7 +356,8 @@ impl Language {
                         .map(|(_, &(ref tp, ref expr))| (tp, expr))
                         .chain(arg_tps.into_iter().zip(xs.iter().map(|&x| x)))
                     {
-                        let free_tp = free_tp.apply(&ctx);
+                        let mut free_tp = free_tp.clone();
+                        free_tp.apply_mut(&ctx);
                         let n = self.likelihood_uses(&free_tp, free_expr, &ctx, env);
                         if n.0.is_infinite() {
                             l = f64::NEG_INFINITY;
@@ -574,13 +576,13 @@ impl<'a> TreeMatcher<'a> {
                 }
             }
             (&Expression::Primitive(f_num), &Expression::Primitive(c_num)) if f_num == c_num => {
-                let tp = &self.dsl.primitives[f_num].1;
-                Some(tp.instantiate(self.ctx))
+                let tp = self.dsl.primitives[f_num].1.clone();
+                Some(tp.instantiate_owned(self.ctx))
             }
             (&Expression::Invented(f_num), &Expression::Invented(c_num)) => {
                 if f_num == c_num {
-                    let tp = &self.dsl.invented[f_num].1;
-                    Some(tp.instantiate(self.ctx))
+                    let tp = self.dsl.invented[f_num].1.clone();
+                    Some(tp.instantiate_owned(self.ctx))
                 } else {
                     None
                 }
@@ -598,7 +600,9 @@ impl<'a> TreeMatcher<'a> {
             (&Expression::Index(i), _) if i < env.len() => {
                 // bound variable
                 if fragment == concrete {
-                    Some(env[i].apply(self.ctx))
+                    let mut tp = env[i].clone();
+                    tp.apply_mut(self.ctx);
+                    Some(tp)
                 } else {
                     None
                 }
