@@ -1,16 +1,22 @@
-use crossbeam_channel::{self, bounded};
-use num_cpus;
 use polytype::{Context, Type, TypeSchema};
-use rayon;
-use rayon::prelude::*;
 use std::collections::VecDeque;
-use std::env;
 use std::f64;
 use std::iter;
 use std::rc::Rc;
 
 use super::{Expression, Language, LinkedList};
 
+#[cfg(feature = "par_enum")]
+use crossbeam_channel::{self, bounded};
+#[cfg(feature = "par_enum")]
+use num_cpus;
+#[cfg(feature = "par_enum")]
+use rayon;
+#[cfg(feature = "par_enum")]
+use rayon::prelude::*;
+#[cfg(feature = "par_enum")]
+use std::env;
+#[cfg(feature = "par_enum")]
 lazy_static! {
     static ref PIECES: u32 = match env::var("RAYON_NUM_THREADS")
         .ok()
@@ -21,6 +27,7 @@ lazy_static! {
     };
 }
 
+#[cfg(feature = "par_enum")]
 const PAR_BUFFER_SIZE: usize = 512;
 const BUDGET_INCREMENT: f64 = 1.0;
 const MAX_DEPTH: u32 = 256;
@@ -41,9 +48,8 @@ pub fn new<'a>(
     Box::new(
         (0..)
             .map(budget_interval)
-            .zip(iter::repeat(tp))
-            .flat_map(move |(budget, tp)| {
-                enumerate(dsl, tp, &ctx, env.clone(), budget, 0)
+            .flat_map(move |budget| {
+                enumerate(dsl, &tp, &ctx, env.clone(), budget, 0)
             })
             // .map(move |(log_prior, log_likelihood, expr)| {
             //     println!("{} {}", log_prior, dsl.display(&expr));
@@ -68,7 +74,7 @@ pub fn new<'a>(
 }
 
 /// enumerate expressions in parallel within the budget interval
-#[cfg_attr(not(feature = "par_enum"), allow(dead_code))]
+#[cfg(feature = "par_enum")]
 fn new_par(
     dsl: Language,
     request: Type,
@@ -81,7 +87,7 @@ fn new_par(
             .zip(exponential_decay(budget))
             .for_each(|(tx, budget)| {
                 let env = Rc::new(LinkedList::default());
-                let e = enumerate(&dsl, request.clone(), &ctx, env.clone(), budget, 0)
+                let e = enumerate(&dsl, &request, &ctx, env.clone(), budget, 0)
                     .map(|(log_prior, _, expr)| (expr, log_prior));
                 for (expr, logprior) in e {
                     if tx.send((expr, logprior)).is_err() {
@@ -94,7 +100,7 @@ fn new_par(
     rx.into_iter()
 }
 
-#[cfg_attr(not(feature = "par_enum"), allow(dead_code))]
+#[cfg(feature = "par_enum")]
 fn exponential_decay(budget: (f64, f64)) -> Vec<(f64, f64)> {
     // because depth values correspond to description length in nats, we
     // assume that for pieces to have the same total description coverage
@@ -172,7 +178,7 @@ fn likelihood_internal<'a>(
 
 fn enumerate<'a>(
     dsl: &'a Language,
-    request: Type,
+    request: &Type,
     ctx: &Context,
     env: Rc<LinkedList<Type>>,
     budget: (f64, f64),
@@ -182,12 +188,12 @@ fn enumerate<'a>(
         Box::new(iter::empty())
     } else if let Some((arg, ret)) = request.as_arrow() {
         let env = LinkedList::prepend(&env, arg.clone());
-        let it = enumerate(dsl, ret.clone(), ctx, env, budget, depth)
+        let it = enumerate(dsl, ret, ctx, env, budget, depth)
             .map(|(ll, ctx, body)| (ll, ctx, Expression::Abstraction(Box::new(body))));
         Box::new(it)
     } else {
         Box::new(
-            dsl.candidates(&request, ctx, &env.as_vecdeque())
+            dsl.candidates(request, ctx, &env.as_vecdeque())
                 .into_iter()
                 .filter(move |&(ll, _, _, _)| -ll <= budget.1)
                 .flat_map(move |(ll, expr, tp, ctx)| {
@@ -214,7 +220,7 @@ fn enumerate_application<'a>(
     if let Some(mut arg_tp) = arg_tps.pop_front() {
         arg_tp.apply_mut(ctx);
         Box::new(
-            enumerate(dsl, arg_tp, ctx, env.clone(), (0f64, budget.1), depth).flat_map(
+            enumerate(dsl, &arg_tp, ctx, env.clone(), (0f64, budget.1), depth).flat_map(
                 move |(arg_ll, ctx, arg)| {
                     let f_next = Expression::Application(Box::new(f.clone()), Box::new(arg));
                     let budget = (budget.0 + arg_ll, budget.1 + arg_ll);
