@@ -6,9 +6,9 @@
 //! # #[macro_use]
 //! # extern crate polytype;
 //! # extern crate programinduction;
-//! use programinduction::pcfg::{task_by_simple_evaluation, Grammar, Rule};
+//! use programinduction::pcfg::{task_by_evaluation, Grammar, Rule};
 //!
-//! fn simple_evaluator(name: &str, inps: &[i32]) -> i32 {
+//! fn evaluator(name: &str, inps: &[i32]) -> i32 {
 //!     match name {
 //!         "0" => 0,
 //!         "1" => 1,
@@ -23,12 +23,12 @@
 //!     vec![
 //!         Rule::new("0", tp!(EXPR), 1.0),
 //!         Rule::new("1", tp!(EXPR), 1.0),
-//!         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
+//!         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
 //!     ],
 //! );
 //!
 //! // task: the number 4
-//! let task = task_by_simple_evaluation(&simple_evaluator, &4, tp!(EXPR));
+//! let task = task_by_evaluation(&evaluator, &4, tp!(EXPR));
 //!
 //! // solution:
 //! let expr = g.parse("plus(1, plus(1, plus(1,1)))").unwrap();
@@ -47,9 +47,9 @@ use std::fmt::Debug;
 use std::sync::Arc;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use itertools::Itertools;
-use polytype::Type;
+use polytype::{Type, TypeSchema};
 use rand::Rng;
-use rand::distributions::{IndependentSample, Range};
+use rand::distributions::Range;
 use rayon::prelude::*;
 
 use {ECFrontier, Task, EC, GP};
@@ -72,8 +72,8 @@ impl Grammar {
     pub fn new(start: Type, all_rules: Vec<Rule>) -> Self {
         let mut rules = HashMap::new();
         for mut rule in all_rules {
-            let nt = if let Type::Arrow(ref arrow) = rule.production {
-                arrow.returns().clone()
+            let nt = if let Some(ret) = rule.production.returns() {
+                ret.clone()
             } else {
                 rule.production.clone()
             };
@@ -100,7 +100,7 @@ impl Grammar {
     ///     vec![
     ///         Rule::new("0", tp!(EXPR), 1.0),
     ///         Rule::new("1", tp!(EXPR), 1.0),
-    ///         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
+    ///         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
     ///     ],
     /// );
     /// let exprs: Vec<AppliedRule> = g.enumerate()
@@ -160,7 +160,7 @@ impl Grammar {
         }
         self.normalize();
     }
-    /// Evaluate a sentence using a simple evaluator.
+    /// Evaluate a sentence using a evaluator.
     ///
     /// # Examples
     ///
@@ -168,9 +168,9 @@ impl Grammar {
     /// # #[macro_use]
     /// # extern crate polytype;
     /// # extern crate programinduction;
-    /// use programinduction::pcfg::{Grammar, Rule, task_by_simple_evaluation};
+    /// use programinduction::pcfg::{Grammar, Rule, task_by_evaluation};
     ///
-    /// fn simple_evaluator(name: &str, inps: &[i32]) -> i32 {
+    /// fn evaluator(name: &str, inps: &[i32]) -> i32 {
     ///     match name {
     ///         "0" => 0,
     ///         "1" => 1,
@@ -185,23 +185,20 @@ impl Grammar {
     ///     vec![
     ///         Rule::new("0", tp!(EXPR), 1.0),
     ///         Rule::new("1", tp!(EXPR), 1.0),
-    ///         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
+    ///         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
     ///     ],
     /// );
     ///
     /// let expr = g.parse("plus(1, plus(1, plus(1,1)))").unwrap();
-    /// assert_eq!(4, g.eval(&expr, &simple_evaluator));
+    /// assert_eq!(4, g.eval(&expr, &evaluator));
     /// # }
     /// ```
-    pub fn eval<V, F>(&self, ar: &AppliedRule, simple_evaluator: &F) -> V
+    pub fn eval<V, F>(&self, ar: &AppliedRule, evaluator: &F) -> V
     where
         F: Fn(&str, &[V]) -> V,
     {
-        let args: Vec<V> = ar.2
-            .iter()
-            .map(|ar| self.eval(ar, simple_evaluator))
-            .collect();
-        simple_evaluator(self.rules[&ar.0][ar.1].name, &args)
+        let args: Vec<V> = ar.2.iter().map(|ar| self.eval(ar, evaluator)).collect();
+        evaluator(self.rules[&ar.0][ar.1].name, &args)
     }
     /// Sample a statement of the PCFG.
     ///
@@ -218,7 +215,7 @@ impl Grammar {
     ///     vec![
     ///         Rule::new("0", tp!(EXPR), 1.0),
     ///         Rule::new("1", tp!(EXPR), 1.0),
-    ///         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
+    ///         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
     ///     ],
     /// );
     /// let ar = g.sample(&tp!(EXPR), &mut rand::thread_rng());
@@ -244,10 +241,10 @@ impl Grammar {
     ///     vec![
     ///         Rule::new("0", tp!(EXPR), 1.0),
     ///         Rule::new("1", tp!(EXPR), 1.0),
-    ///         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
-    ///         Rule::new("zero?", arrow![tp!(EXPR), tp!(BOOL)], 1.0),
-    ///         Rule::new("if", arrow![tp!(BOOL), tp!(EXPR), tp!(EXPR)], 1.0),
-    ///         Rule::new("nand", arrow![tp!(BOOL), tp!(BOOL), tp!(BOOL)], 1.0),
+    ///         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
+    ///         Rule::new("zero?", tp!(@arrow[tp!(EXPR), tp!(BOOL)]), 1.0),
+    ///         Rule::new("if", tp!(@arrow[tp!(BOOL), tp!(EXPR), tp!(EXPR)]), 1.0),
+    ///         Rule::new("nand", tp!(@arrow[tp!(BOOL), tp!(BOOL), tp!(BOOL)]), 1.0),
     ///     ]
     /// );
     ///
@@ -283,7 +280,7 @@ impl Grammar {
     /// [`parse`]: #method.parse
     pub fn display(&self, ar: &AppliedRule) -> String {
         let r = &self.rules[&ar.0][ar.1];
-        if let Type::Arrow(_) = r.production {
+        if r.production.as_arrow().is_some() {
             let args = ar.2.iter().map(|ar| self.display(ar)).join(",");
             format!("{}({})", r.name, args)
         } else {
@@ -329,8 +326,14 @@ impl EC for Grammar {
     type Expression = AppliedRule;
     type Params = EstimationParams;
 
-    fn enumerate<'a>(&'a self, tp: Type) -> Box<Iterator<Item = (Self::Expression, f64)> + 'a> {
-        self.enumerate_nonterminal(tp)
+    fn enumerate<'a>(
+        &'a self,
+        tp: TypeSchema,
+    ) -> Box<Iterator<Item = (Self::Expression, f64)> + 'a> {
+        match tp {
+            TypeSchema::Monotype(tp) => self.enumerate_nonterminal(tp),
+            _ => panic!("PCFGs can't handle polytypes"),
+        }
     }
     /// This is exactly the same as [`Grammar::update_parameters`], but optimized to deal with
     /// frontiers.
@@ -393,8 +396,12 @@ impl GP for Grammar {
         _params: &Self::Params,
         rng: &mut R,
         pop_size: usize,
-        tp: &Type,
+        tp: &TypeSchema,
     ) -> Vec<Self::Expression> {
+        let tp = match *tp {
+            TypeSchema::Monotype(ref tp) => tp,
+            _ => panic!("PCFGs can't handle polytypes"),
+        };
         (0..pop_size).map(|_| self.sample(tp, rng)).collect()
     }
     fn mutate<R: Rng>(
@@ -404,8 +411,7 @@ impl GP for Grammar {
         prog: &Self::Expression,
     ) -> Self::Expression {
         let tot = params.mutation_point + params.mutation_subtree + params.mutation_reproduction;
-        let _ = rng;
-        match Range::new(0f64, tot).ind_sample(rng) {
+        match Range::sample_single(0f64, tot, rng) {
             x if x < params.mutation_point => mutate_random_node(prog.clone(), rng, |ar, rng| {
                 let rule = &self.rules[&ar.0][ar.1];
                 let mut candidates: Vec<_> = self.rules[&ar.0]
@@ -499,7 +505,7 @@ fn update_counts<'a>(ar: &'a AppliedRule, counts: &Arc<HashMap<Type, Vec<AtomicU
 /// `V` will often be an enum corresponding to each nonterminal in the PCFG. All outputs and
 /// evaluated sentences must be representable by `V`.
 ///
-/// A `simple_evaluator` takes the name of a production and a vector corresponding to evaluated results
+/// An `evaluator` takes the name of a production and a vector corresponding to evaluated results
 /// of each child node of the production in a particular derivation.
 ///
 /// The resulting task is "all-or-nothing": the oracle returns either `0` if all examples are
@@ -511,9 +517,9 @@ fn update_counts<'a>(ar: &'a AppliedRule, counts: &Arc<HashMap<Type, Vec<AtomicU
 /// # #[macro_use]
 /// # extern crate polytype;
 /// # extern crate programinduction;
-/// use programinduction::pcfg::{Grammar, Rule, task_by_simple_evaluation};
+/// use programinduction::pcfg::{task_by_evaluation, Grammar, Rule};
 ///
-/// fn simple_evaluator(name: &str, inps: &[i32]) -> i32 {
+/// fn evaluator(name: &str, inps: &[i32]) -> i32 {
 ///     match name {
 ///         "0" => 0,
 ///         "1" => 1,
@@ -528,20 +534,20 @@ fn update_counts<'a>(ar: &'a AppliedRule, counts: &Arc<HashMap<Type, Vec<AtomicU
 ///     vec![
 ///         Rule::new("0", tp!(EXPR), 1.0),
 ///         Rule::new("1", tp!(EXPR), 1.0),
-///         Rule::new("plus", arrow![tp!(EXPR), tp!(EXPR), tp!(EXPR)], 1.0),
+///         Rule::new("plus", tp!(@arrow[tp!(EXPR), tp!(EXPR), tp!(EXPR)]), 1.0),
 ///     ],
 /// );
 ///
 /// let output = 4;
 /// let tp = tp!(EXPR);
-/// let task = task_by_simple_evaluation(&simple_evaluator, &output, tp);
+/// let task = task_by_evaluation(&evaluator, &output, tp);
 ///
 /// let expr = g.parse("plus(1, plus(1, plus(1,1)))").unwrap();
 /// assert!((task.oracle)(&g, &expr).is_finite())
 /// # }
 /// ```
-pub fn task_by_simple_evaluation<'a, V, F>(
-    simple_evaluator: &'a F,
+pub fn task_by_evaluation<'a, V, F>(
+    evaluator: &'a F,
     output: &'a V,
     tp: Type,
 ) -> Task<'a, Grammar, AppliedRule, &'a V>
@@ -550,7 +556,7 @@ where
     F: Fn(&str, &[V]) -> V + Sync + 'a,
 {
     let oracle = Box::new(move |g: &Grammar, ar: &AppliedRule| {
-        if output == &g.eval(ar, simple_evaluator) {
+        if output == &g.eval(ar, evaluator) {
             0f64
         } else {
             f64::NEG_INFINITY
@@ -559,7 +565,7 @@ where
     Task {
         oracle,
         observation: output,
-        tp,
+        tp: TypeSchema::Monotype(tp),
     }
 }
 
